/*********************************************************************
 * Desktop Station air Sketch for DSwio Hardware Platform
 *
 * Copyright (C) 2020 DesktopStation Co.,Ltd. / Yaasan
 * https://desktopstation.net/
 */

#define DEBUG

// include the library code:
#include <TFT_eSPI.h> // Graphics and font library for ILI9341 driver chip
#include <SPI.h>
#include "DSCoreM_Type.h"
#include "DSCoreM_Common.h"
#include "DSCoreM_List.h"
#include "DSCoreM_DCC.h"
#include "DSCoreM_MM2.h"
#include "DSCoreM.h"
#include "Functions.h"
#include "TrackReporterS88_DS.h"

#define FIRMWARE_VER   '5' /* 0-9, a-z, A-Z*/

#define MAX_S88DECODER 1

#define LEDSTATE_STOP 0
#define LEDSTATE_RUN 1
#define LEDSTATE_ERR 2
#define LEDSTATE_NOSD 3
#define LEDSTATE_ANALOG 4
#define PIN_PWMA 9
#define PIN_PWMB 10
#define PIN_EDC A1
#define PIN_RUNLED	2
#define PIN_ALERT 3
#define PIN_S88CHK A0
#define PIN_SPICS 4
#define PIN_CURRENT A2

#define SMEM_POWER		0
#define SMEM_ERROR		1
#define SMEM_EDC		2
#define SMEM_CUR		3
#define SMEM_SEQ		4
#define SMEM_S88_1		5
#define SMEM_S88_2		6
#define SMEM_S88_3		7
#define SMEM_S88_4		8

#define THRESHOLD_EDC 276  // 1024d * (1.2 / 8) * 9V / 5V
#define THRESHOLD_EDC_DSA 375  // 1024d * (1.2 / 5.9) * 9V / 5V
#define THRESHOLD_OV  553  // 1024d * (1.2 / 8) * 20V / 5V
#define THRESHOLD_OV_DSA  999  // 1024d * (1.2 / 5.9) * 20V / 5V

#define TIMEOUT_SERIAL 10

#define MAX_LOCSTATUS_WEB	8

#define SHAREMEM_ACCSIZE 64
#define SHAREMEM_SIZE 32

#define ADC_SHIFT 0

#define TFT_GREY 0x0000 // New colour


typedef struct {
	byte mOV:1;
	byte mOC:1;
	byte mLV:1;
	byte mUnk1:1;
	byte mUnk2:1;
	byte reserved:3;
} DATA_STATUS;

typedef struct {
	word mAddress;//Marklin Address Format
	byte mSpeed;//(1/4 Data)
	byte mDir;
	word mFuncBuf1;
	word mFuncBuf2;
} DATA_LOC;



/* 機関車データ確認 */
DATA_LOC gLocStatus[MAX_LOCSTATUS_WEB];

/* ポイントのデータ */
uint8_t buffer_accessories[SHAREMEM_ACCSIZE];


/* Decralation classes */
DSCoreLib DSCore;
TrackReporterS88_DS reporter(MAX_S88DECODER);


/* LED functions */
byte gLED_State = LEDSTATE_STOP;
byte gLED_Toggle = 0;
byte gLED_Counter = 0;

/* Sensor */
short gOffset_Current = 0;
word gEdc_data = 0;
short gCur_data = 0;
uint8_t gSwSensor = 0;
uint8_t gMaxS88Num = 0;

/* Error Threshold */
word gThreshold_OV = 0;
word gThreshold_LV = 0;
byte gLastErrorNo = 0;
uint8_t gAliveSeq = 0;

/* Parser */
String gRequest;
String function;
word arguments[4];
byte numOfArguments;
word gLocAddresses[4] = {0,0,0,0};
byte numOfLocAddr = 0;
boolean result;

DATA_STATUS gStatus;

//TFT LCD

TFT_eSPI tft = TFT_eSPI();  // Invoke library


//Task Schedule
unsigned long gPreviousL7 = 0; // 1000ms interval(Packet Task)
unsigned long gPreviousL5 = 0; //   50ms interval
unsigned long gPreviousL8 = 0; // 1000ms interval


void SetAnalogSpeed(word inSpeed, byte inDir);
void printString(const char *s, int x, int y);
void ReplyPowerPacket(byte inPower);


void LOCMNG_Clear(void);
byte LOCMNG_SetLocoSpeed(word inAddr, word inSpd);
byte LOCMNG_SetLocoDir(word inAddr, byte inDir);
byte LOCMNG_SetLocoFunc(word inAddr, byte inNo, byte inPower);
void LOCMNG_UpdateAcc(word inNo);

/***********************************************************************
* 
* Boot setup
* 
* 
************************************************************************/

void setup()
{
    tft.init();
    tft.setRotation(1);
	
	tft.fillScreen(TFT_GREY);
	
	tft.setCursor(1, 0, 2);
	tft.setTextColor(TFT_WHITE, TFT_BLACK);
	tft.setTextSize(2);
	tft.println("DesktopStation DSwio");
	tft.println("Welcome onboard!");
	
	/* Debug message for PC */
	Serial.begin(115200);
	
	Serial.println("Desktop Station air");
	tft.println("- Serial");

	//初期化
	LOCMNG_Clear();
	
	for( int i = 0; i < SHAREMEM_ACCSIZE; i++)
	{
		buffer_accessories[i] = 0;
	}
	
	//Pin mode
	pinMode(PIN_EDC, INPUT);//Edc
	pinMode(SENSOR_CURRENT, INPUT);
	pinMode(PIN_RUNLED, OUTPUT);
	pinMode(PIN_ALERT, INPUT);
	digitalWrite(PIN_ALERT, HIGH);
	
	pinMode(PIN_S88CHK, INPUT);
	digitalWrite(PIN_S88CHK, HIGH);
	
	tft.println("- Pins");

	/* DSCore初期化 */
	DSCore.Init();
	
	tft.println("- DSCore");
	
	/* 線路に電源遮断 */
	DS_Power(0);
	
	/* Dammy sleep */
	CalcCurrentOffset();
	
	
	/* PULLUPを解除 */
	digitalWrite(PIN_S88CHK, LOW);
	
	Serial.println("-------------");
	tft.println("Ready.");

	
	/* LED turn on */
	changeLED();
	
	Serial.println(F("100 Ready"));
	
	//Reset task
	gPreviousL5 = millis();
	gPreviousL8 = millis();
	gPreviousL7 = millis();

	//Timer1 490Hz -> 31.4kHz
	//TCCR1B &= B11111000;
	//TCCR1B |= B00000001;	
	
	
}


boolean parse() 
{
	int aPosSlash = -1;

	char aTemp[SHAREMEM_SIZE];
	gRequest.toCharArray(aTemp, SHAREMEM_SIZE);

	int lpar = gRequest.indexOf('(');
	if (lpar == -1) 
	{
		return false;
	}

	function = String(gRequest.substring(0, lpar));
	function.trim();

	int offset = lpar + 1;
	int aAdddrOffset = 0;
	int comma = gRequest.indexOf(',', offset);
	numOfArguments = 0;
	numOfLocAddr = 0;
	
	while (comma != -1) 
	{
		String tmp = gRequest.substring(offset, comma);
		tmp.trim();
		
		/* スラッシュを探索する */
		aPosSlash = tmp.indexOf('/', 0);
		
		if( aPosSlash > 0)
		{
			tmp = tmp + "/";
			
			//Dammy値を指定
			arguments[numOfArguments++] = 0;
			aAdddrOffset = 0;
			
			//スラッシュはアドレス部だけに許可
			while( aPosSlash != -1)
			{
				String tmp2 = tmp.substring(aAdddrOffset, aPosSlash);
				tmp2.trim();
				
				if( numOfLocAddr < 4)
				{
					gLocAddresses[numOfLocAddr] = stringToWord(tmp2);
					numOfLocAddr++;
				}
				
				aAdddrOffset = aPosSlash + 1;
				aPosSlash = tmp.indexOf('/', aAdddrOffset);
			}
			
			offset = comma + 1;
			comma = gRequest.indexOf(',', offset);
		}
		else
		{
			arguments[numOfArguments++] = stringToWord(tmp);
			offset = comma + 1;
			comma = gRequest.indexOf(',', offset);
		}
	}

	int rpar = gRequest.indexOf(')', offset);
	while (rpar == -1) 
	{
		return false;
	}

	if (rpar > offset) 
	{
		String tmp = gRequest.substring(offset, rpar);
		tmp.trim();
		arguments[numOfArguments++] = stringToWord(tmp);
	}

	return true;
}

boolean dispatch() 
{
	byte aValue;
	boolean aResult;
	byte aDir = 0;
	
	if ((function == "setLocoDirection") || ((function == "DI")))
	{
		
		if( arguments[1] > 0)
		{
			//FWD=1,REV=2を FWD=0,REV=1に変換
			aDir = arguments[1] - 1;
			
			//異常値はREV固定
			if(aDir > 1)
			{
				aDir = 1;
			}
		}
		else
		{
			//FWD固定
			aDir = 0;
		}
		
		if( numOfLocAddr > 0)
		{
			//重連対応
			for( int i = 0; i < numOfLocAddr; i++)
			{
				LOCMNG_SetLocoDir(gLocAddresses[i], aDir);
				aResult =  DSCore.SetLocoDirection(gLocAddresses[i], aDir);
			}
			
			aResult = true;
		}
		else
		{
			//非重連
			LOCMNG_SetLocoDir(arguments[0], aDir);
			aResult =  DSCore.SetLocoDirection(arguments[0], aDir);
		}
		
		return aResult;
	}
	else if ((function == "setLocoFunction") || ((function == "FN")))
	{
		LOCMNG_SetLocoFunc(arguments[0], arguments[1], arguments[2]);
		
		return DSCore.SetLocoFunction(arguments[0], arguments[1] + 1, arguments[2]);
	
	}
	else if ((function == "setTurnout") || ((function == "TO")))
	{
		//テーブル更新
		//MM2: 1(0x3000) - 320(0x3140)
		//DCC: 1(0x3800) - 2044(0x3FFC)
		
		word aAccAddr = 0;
		
		if( (arguments[0] >= 0x3000) && (arguments[0] <= 0x3140))
		{
			aAccAddr = arguments[0] - 0x3000;
		}
		else if( (arguments[0] >= 0x3800) && (arguments[0] <= 0x3FFC))
		{
			aAccAddr = arguments[0] - 0x3800;
		}
		
		if( (aAccAddr >> 3) <= 31)
		{
			if( arguments[1] == 0)
			{
				buffer_accessories[aAccAddr >> 3] = buffer_accessories[aAccAddr >> 3] & ~(1UL << (aAccAddr & 0x07));
			}
			else
			{
				buffer_accessories[aAccAddr >> 3] = buffer_accessories[aAccAddr >> 3] | (1UL << (aAccAddr & 0x07));
			}
			
		}
		
		//ポイントパケット出力登録
		
		return DSCore.SetTurnout(arguments[0], (byte)arguments[1]);
	
	}
	else if ((function == "setPower") || ((function == "PW")))
	{
		
		if( arguments[0] == 1)
		{
			//端子を解放
			FreeAnalogSpeed();
			
			//電流オフセット計測
			CalcCurrentOffset();
			
			/* 状態変更 */
			gLED_State = LEDSTATE_RUN;
			
		}
		else
		{
			/* 状態変更 */
			gLED_State = LEDSTATE_STOP;
		}
		
		
		/* LED状態の変更 */
		gLED_Counter = 0;
		changeLED();
		
		return DSCore.SetPower(arguments[0]);
	
	}
	else if ((function == "setLocoSpeed") || ((function == "SP")))
	{
		if( numOfLocAddr > 0)
		{
			//重連対応
			for( int i = 0; i < numOfLocAddr; i++)
			{
				LOCMNG_SetLocoSpeed(gLocAddresses[i], arguments[1]);
				aResult =  DSCore.SetLocoSpeedEx(gLocAddresses[i], arguments[1], arguments[2]);
			}
			aResult = true;
		}
		else
		{
			//非重連
			LOCMNG_SetLocoSpeed(arguments[0], arguments[1]);
			aResult =  DSCore.SetLocoSpeedEx(arguments[0], arguments[1], arguments[2]);
		}
		
		return aResult;
	}
	else if (function == "DC")
	{
		//PRM0=Speed(0-1023), PRM1=Dir(0,1=FWD, 2=REV)
		if( (gLED_State == LEDSTATE_STOP) || (gLED_State == LEDSTATE_ANALOG))
		{
			
			if( arguments[0] > 0)
			{
				gLED_State = LEDSTATE_ANALOG;
			
				SetAnalogSpeed(arguments[0], arguments[1]);
			}
			else
			{
				SetAnalogSpeed(0, 0);
				gLED_State = LEDSTATE_STOP;
			}
		}
		
		return true;
	}
	else if ((function == "reset") || ((function == "RS")))
	{
		#ifdef DEBUG
		Serial.println(F("100 Ready"));
		#endif
		
		/* リセットパケットを投げる */
		if( gLED_State == LEDSTATE_RUN)
		{
			//内部のデータバッファを削除する
			DSCore.Clear();
			
			//リセット
			DSCore.SendReset();
			
		}
		
		return true;
	}
	/* reset */
	else if ((function == "setPing") || ((function == "PG")))
	{
		/* Type(2bytes), Version(2bytes), UID(4bytes) */
		Serial.println(F("@DSG,00,03,00,02,00,00,00,00"));
		return true;
	}
	else if ((function == "getLocoConfig") || ((function == "GV")))
	{
		
		/* LED turn on */
		digitalWrite(PIN_RUNLED, HIGH);
		

		aValue = 0;
		aResult = DSCore.ReadConfig( arguments[1], &aValue, 0, 0);
		
		
		//Send to Serial
		Serial.print("@CV,");
		Serial.print(arguments[0]);
		Serial.print(",");
		Serial.print(arguments[1]);
		Serial.print(",");
		
		if( aResult == true)
		{
			Serial.print(aValue);
		}
		else
		{
			Serial.print(0xFFFF);
		}
		
		Serial.println(",");
		/* LED turn off */
		changeLED();
		
		return true;
	}
	else if ((function == "getS88") || (function == "gS8"))
	{
		int aMaxS88Num = MAX_S88DECODER;
		
		if( arguments[0] > 0)
		{
			aMaxS88Num = arguments[0];
		}
		
		//Send a S88 sensor reply 
		Serial.print("@S88,");

		reporter.refresh(aMaxS88Num);

		word aFlags = 0;

		for( int j = 0; j < aMaxS88Num; j++)
		{
		  aFlags = (reporter.getByte((j << 1) + 1) << 8) + reporter.getByte(j << 1);
		  
		  Serial.print((word)aFlags, HEX);
		  Serial.print(",");
		}
			
		Serial.println("");
		
		return true;
		
	} /* getS88 */
	else if ((function == "setLocoConfig") || ((function == "SV")))
	{
		
		Serial.print("@CVWrite,");
		Serial.print(arguments[1]);
		Serial.print(",");
		Serial.print(arguments[2]);
		Serial.println("");
		
		
		/* LED turn on */
		digitalWrite(PIN_RUNLED, HIGH);
		
		
		DSCore.WriteConfig_Dir( arguments[1], arguments[2]);
		
	
		/* LED turn off */
		changeLED();

		return true;
	}
	else if ((function == "cutout") || ((function == "CO")))
	{
		/* Railcom CutOut */
		DSCore.gCutOut = (arguments[0] == 1) ? 1: 0;
		
		return true;
	}

	else
	{
		return false;
	}
}

/***********************************************************************
* 
* Timer Interval
* 
************************************************************************/

void loop()
{
	long aEdc;
	signed long aCur;
	uint8_t aErrNo = 1;
	
	
	//パルスジェネレータのスキャン
	DSCore.Scan();
	
	if( (millis() - gPreviousL5) >= 200)
	{
		//Reset task
		gPreviousL5 = millis();
		
		//LED Control
		taskLED();
		
	}
	
	if( (millis() - gPreviousL8) >= 500)
	{
		
		//Reset task
		gPreviousL8 = millis();
		
		//S88 Scan
		ScanS88();
		
		/* Sensor check */
		if( gSwSensor == 0)
		{
			gSwSensor = 1;
			gEdc_data = analogRead(PIN_EDC) >> ADC_SHIFT;
			
			/* OV check */
			if( gEdc_data > gThreshold_OV )
			{
				aErrNo = 4;
			}
			
			/* LV check */
			if( gEdc_data < gThreshold_LV )
			{
				aErrNo = 2;
				
			}
			
			// 1024d * (1.2 / 5.9) * Y[V] / 5[V] = X
			// -> 
			aEdc = (((long)gEdc_data * 246) >> 10);
			
			if( aEdc > 255)
			{
				aEdc = 255;
			}
			
		}
		else
		{
			gSwSensor = 0;
			gCur_data = analogRead(PIN_CURRENT) >> ADC_SHIFT;
			
			/* 電流の向きを正規化 */
			if( gCur_data >= gOffset_Current)
			{
				gCur_data = gCur_data - gOffset_Current;
			}
			else
			{
				gCur_data = gOffset_Current - gCur_data;
			}			
			
			aCur = ((long)gCur_data * 651) / 213;
			
			if( aCur > 1250)
			{
				aCur = 1250;
			}
			
			
		}
		
		
		/* DSairR2 HW  */
		if( (digitalRead(PIN_ALERT) == 1) )
		{
			//Over current at TB6642FG
			aErrNo = 8;
		}
		
		
		/* エラー処理 */
		if( aErrNo > 1)
		{
			if( DSCore.IsPower() == 1)
			{
				//PowerOffByErr(aErrNo);
				//Serial.println(aErrNo);
			}
			
			
		}

		
		//パルスジェネレータのスキャン
		DSCore.Scan();
	}
	
		
		int aReceived = receiveRequest();
		
		if( aReceived > 0)
		{
			
	  		if (parse()) {
				
	  			#ifdef DEBUG
				Serial.println(gRequest);
				#endif
	  			
				if (dispatch()) {
					Reply200();
				} else {
					Reply300();
				}
	  			
			} else {
				Reply301();
			}
			
			/* Reply to Desktop Station */
		}
		else
		{
			/* Nothing to do */
			
		}		
		

}

void Reply200(void)
{
	Serial.println(F("200 Ok"));
}

void Reply300(void)
{
	Serial.println(F("300 Cmd err"));
}

void Reply301(void)
{
	Serial.println(F("301 Stx err"));
}

void Reply302(void)
{
	Serial.println(F("302 Timeout"));
}

void Reply303(void)
{
	Serial.println(F("303 Unkn err"));
}

void Reply304(void)
{
	Serial.println(F("304 Ser err"));
}


void ScanS88(void)
{
	if( gMaxS88Num > 0)
	{
		reporter.refresh(gMaxS88Num);
		
	}
}



void PowerOffByErr(uint8_t inErrNo)
{
	DSCore.SetPower(0);
	delay(10);
	ReplyPowerPacket(0);
	DSCore.SetPower(0);
	
	/* 状態変更 */
	gLED_State = LEDSTATE_ERR;
	gLED_Counter =0;
	gLastErrorNo = inErrNo;
	
	
}

void ReplyPowerPacket(byte inPower)
{
	#ifdef DEBUG
	/* Power */
	Serial.print("@PWR,");
	Serial.print(inPower);
	Serial.println(",");
	#endif
}


void changeLED(void)
{
	
	switch( gLED_State)
	{
	case LEDSTATE_STOP:
		digitalWrite(PIN_RUNLED, LOW);
		break;
		
	case LEDSTATE_RUN:
		digitalWrite(PIN_RUNLED, HIGH);
		break;
	
	case LEDSTATE_ANALOG:
		break;
		
	case LEDSTATE_ERR:
		break;
		
	case LEDSTATE_NOSD:
		break;
	}
}

void taskLED(void)
{
	
	/* カウンタを回す(200msで1カウント) */
	gLED_Counter++;
	gLED_Counter = gLED_Counter & 0b1111;
	
	//LED操作
	switch( gLED_State)
	{
	case LEDSTATE_STOP:
		
		if( gLED_Counter <= 3)
		{
			digitalWrite(PIN_RUNLED, HIGH);
		}
		else
		{
			digitalWrite(PIN_RUNLED, LOW);
		}
		break;
		
	case LEDSTATE_RUN:
		break;
		
	case LEDSTATE_ERR:
		
		if( gLED_Toggle == 0)
		{
			digitalWrite(PIN_RUNLED, LOW);
		}
		else
		{
			digitalWrite(PIN_RUNLED, HIGH);
		}
		
		gLED_Toggle = (~gLED_Toggle) & 1;
		break;
	
	case LEDSTATE_ANALOG:
		if( (gLED_Counter == 0) || (gLED_Counter == 8))
		{
			digitalWrite(PIN_RUNLED, LOW);
		}
		else
		{
			digitalWrite(PIN_RUNLED, HIGH);
		}
		
		break;
		
	case LEDSTATE_NOSD:
		
		if( gLED_Counter <= 0)
		{
			digitalWrite(PIN_RUNLED, HIGH);
		}
		else if(gLED_Counter == 2)
		{
			digitalWrite(PIN_RUNLED, HIGH);
		}
		else
		{
			digitalWrite(PIN_RUNLED, LOW);
		}
		
		break;
	}
}

void SetAnalogSpeed(word inSpeed, byte inDir)
{
	
	word aSpeed = inSpeed >> 2;
	
	if( aSpeed > 255)
	{
		aSpeed = 255;
	}
	
	if( inSpeed == 0)
	{
		FreeAnalogSpeed();
	}
	else
	{
		if( inDir == 2)
		{
			//REV
			analogWrite(PIN_PWMA, 255 - aSpeed);
			analogWrite(PIN_PWMB, 255);
		}
		else
		{
			//FWD
			analogWrite(PIN_PWMB, 255 - aSpeed);
			analogWrite(PIN_PWMA, 255);
		}
	}
}

void FreeAnalogSpeed()
{
	analogWrite(PIN_PWMB, 0);
	analogWrite(PIN_PWMA, 0);
}


int receiveRequest() {
	char buffer[64];
	int i = 0;
	char aByte;
	int aResult = -1;
	
	unsigned long time = millis();
	
	/* Check serial buffer */
	if (Serial.available() > 0) {
		
		while (1) {
		
			if (Serial.available() > 0) {
			
				aByte = Serial.read();
				
				// Write to text buf.
				buffer[i] = aByte;
				
				// check last identification of text end.
				if (aByte == '\n') {
					aResult = i;
					break;
				}
				
				// increment
				i = i + 1;
				
				// buf over check
				if( i > 64)
				{
					break;
				}
			}
			else
			{
				/* Check timeout. */
				if( millis() > time + 1000)
				{
					aResult = -1;
					break;
				}
			}
		}
	}
	else
	{
		/* Not received. */
		aResult = 0;
	}
	
	if( aResult > 0)
	{
	
		buffer[i] = '\0';
		
		gRequest = String(buffer);
	}
	else
	{
		gRequest = "";
	}
	
	return aResult;
}

void changedTargetSpeed(const word inAddr, const word inSpeed, const bool inUpdate)
{
	byte aProtocol = 2;
	
	if( inAddr <= 255)
	{
		aProtocol = 0;
	}
	
	if( inUpdate == true)
	{
		DSCore.SetLocoSpeedEx(inAddr, inSpeed, aProtocol);
	}
}

void changedTargetFunc(const word inAddr, const byte inFuncNo, const byte inFuncPower, const bool inUpdate)
{
	if( inUpdate == true)
	{
		DSCore.SetLocoFunction(inAddr, inFuncNo, inFuncPower);
	}

}

void changedTargetDir(const word inAddr, const byte inDir, const bool inUpdate)
{
	
	if( inUpdate == true)
	{
		DSCore.SetLocoDirection(inAddr, inDir);
	}
}

void changedTurnout(const word inAddr, const byte inDir, const bool inUpdate)
{
	if( inUpdate == true)
	{
		DSCore.SetTurnout(inAddr, inDir);
	}
	
}


void changedPowerStatus(const byte inPower)
{
	
	DSCore.SetPower(inPower);
	
}


void CalcCurrentOffset()
{
	unsigned short aCurrentSum = 0;
	
	if( gOffset_Current == 0)
	{
		/* 直流中間に電圧が印加されてないと、実際の動作時のADCのオフセット誤差が異なる */
		if( analogRead(PIN_EDC) >= THRESHOLD_EDC)
		{
			
			for( int i = 0; i < 4; i++)
			{
				aCurrentSum = aCurrentSum + (analogRead(PIN_CURRENT) >> ADC_SHIFT);
				delay(10);
			}
			
			//平均値を算出
			gOffset_Current = aCurrentSum >> 2;
		}
	}
	
	/* 検出できない場合は512をオフセット値にセット */
	if( gOffset_Current == 0)
	{
		gOffset_Current = 512;
	}
}

void LOCMNG_Clear(void)
{
	for( int i = 0; i < MAX_LOCSTATUS_WEB; i++)
	{
		gLocStatus[i].mAddress = 0;
		gLocStatus[i].mSpeed = 0;
		gLocStatus[i].mDir = 0;
		gLocStatus[i].mFuncBuf1 = 0;
		gLocStatus[i].mFuncBuf2 = 0;
	}
}


byte LOCMNG_SetLocoSpeed(word inAddr, word inSpd)
{
	byte aIndex = 255;
	word aSpeed = 0;
	
	for( int i = 0; i < MAX_LOCSTATUS_WEB; i++)
	{
		if( (gLocStatus[i].mAddress == inAddr) || ( gLocStatus[i].mAddress == 0 ))
		{
			aIndex = i;
			break;
		}
	}
	
	if( aIndex != 255)
	{
		gLocStatus[aIndex].mAddress = inAddr;
		aSpeed = inSpd >> 2;
		
		if( aSpeed > 255)
		{
			aSpeed = 255;
		}
		
		char aStrAddr[5];
		char aStrSpd[3];
		
		sprintf(aStrAddr, "%04X", gLocStatus[aIndex].mAddress);
		sprintf(aStrSpd, "%02X", aSpeed);
		
		/* 共有メモリ書き換え */
		
	}
	
	return 0;
}

byte LOCMNG_SetLocoDir(word inAddr, byte inDir)
{
	byte aIndex = 255;
	
	for( int i = 0; i < MAX_LOCSTATUS_WEB; i++)
	{
		if( (gLocStatus[i].mAddress == inAddr) || ( gLocStatus[i].mAddress == 0 ))
		{
			aIndex = i;
			break;
		}
	}
	
	if( aIndex != 255)
	{
		gLocStatus[aIndex].mAddress = inAddr;
		
		if( gLocStatus[aIndex].mDir != inDir)
		{
			gLocStatus[aIndex].mDir = inDir;
			
		}
		
		char aStrAddr[5];
		char aStrDir;
		
		sprintf(aStrAddr, "%04X", gLocStatus[aIndex].mAddress);
		aStrDir = (gLocStatus[aIndex].mDir == 0) ? '0' : '1';
		/* 共有メモリ書き換え */
	}
	
	return 0;
}

byte LOCMNG_SetLocoFunc(word inAddr, byte inNo, byte inPower)
{
	char aStrAddr[5];
	char aStrFunc1[5];
	char aStrFunc2[5];
	byte aIndex = 255;
	
	for( int i = 0; i < MAX_LOCSTATUS_WEB; i++)
	{
		if( (gLocStatus[i].mAddress == inAddr) || ( gLocStatus[i].mAddress == 0 ))
		{
			aIndex = i;
			break;
		}
	}
	
	if( aIndex != 255)
	{
		gLocStatus[aIndex].mAddress = inAddr;
		sprintf(aStrAddr, "%04X", gLocStatus[aIndex].mAddress);

		
		if( inPower == 0)
		{
			if( inNo < 16)
			{
				gLocStatus[aIndex].mFuncBuf1 = gLocStatus[aIndex].mFuncBuf1 & ~(1 << inNo);
			}
			else
			{
				gLocStatus[aIndex].mFuncBuf2 = gLocStatus[aIndex].mFuncBuf2 & ~(1 << (inNo - 16));
			}
		}
		else
		{
			if( inNo < 16)
			{
				gLocStatus[aIndex].mFuncBuf1 = gLocStatus[aIndex].mFuncBuf1 | (1 << inNo);
			}
			else
			{
				gLocStatus[aIndex].mFuncBuf2 = gLocStatus[aIndex].mFuncBuf2 | (1 << (inNo - 16));
			}
		}
		
		
		sprintf(aStrFunc1, "%04X", gLocStatus[aIndex].mFuncBuf1);
		sprintf(aStrFunc2, "%04X", gLocStatus[aIndex].mFuncBuf2);
		
		
		/* FlashAir共有メモリ書き換え */
		
		
	}
	
	return 0;
}